import discord
from discord.ext import commands

from tagger.visitor.obj_builtin import *

class UserBase(TagBase):
    NAME = 'User'

class User(TagInstance):
    
    def __init__(self, user: discord.User | discord.Member) -> None:
        super().__init__()
        self.user = user
        self.set_type(user_)
        
        self.field['name'] = TagString(user.name)
        self.field['id'] = TagInteger(user.id)
        self.field['avatar'] = TagString(user.display_avatar.url)
        self.field['bot'] = TagInteger(int(user.bot))

class MessageBase(TagBase):
    NAME = 'Message'

class Message(TagInstance):
    
    def __init__(self, msg: discord.Message) -> None:
        super().__init__()
        self.msg = msg
        self.set_type(message)
        
        self.field['author'] = User(self.msg.author)
        self.field['id'] = TagInteger(self.msg.id)
        self.field['content'] = TagString(self.msg.content)

class ContextTagBase(TagBase):
    NAME = 'MessageTag'

class ContextTag(TagInstance):
    
    def __init__(self, bot: commands.Bot, message: discord.Message, ctx: commands.Context) -> None:
        super().__init__()
        self.msg = message
        self.bot = bot
        self.ctx = ctx
        self.set_type(messagetag)
        
        self.field['user'] = User(self.ctx.message.author)
        self.field['message'] = Message(message)
        
        add_Amethod(self, 'set', _ctx_set)
        add_Amethod(self, 'add', _ctx_add)
        add_Amethod(self, 'wait_for_react', _ctx_waitfor_reaction)
        add_Amethod(self, 'wait_for_reply', _ctx_waitfor_reply)

messagetag = ContextTagBase()
user_ = UserBase()
message = MessageBase()

async def get_this_msg(env: Enviroment) -> ContextTag:
    return typing.cast(ContextTag, env.get('this'))

async def _ctx_set(args: TagArg, kwargs: Kargs, env: Enviroment) -> TagInstance:
    this = await get_this_msg(env)
    
    try:
        await this.msg.edit(content=(' '.join([arg.__repr__() for arg in args] or ['``````'])))
    except Exception as e:
        return make_error('DisTagError', e.args[0].__str__())
    
    return tag_null

async def _ctx_add(args: TagArg, kwargs: Kargs, env: Enviroment):
    this = await get_this_msg(env)
    
    try:
        await this.msg.edit(content=(this.msg.content + (' '.join([arg.__repr__() for arg in args] or ['']))))
    except Exception as e:
        return make_error('DisTagError', e.args[0].__str__())
    
    return tag_null

async def _ctx_waitfor_reaction(args: TagArg, kwargs: Kargs, env: Enviroment):
    this = await get_this_msg(env)
    
    success, er = expect('MessageTag.wait_for_react', [TagBaseString], {}, args, kwargs)
    if not success:
        return er
    
    t = tstring(args[0])
    
    r: discord.Reaction | None = this.bot.wait_for('reaction_add', check=lambda r: r.emoji.name == t, timeout=100)
    
    if r is None:
        return tag_null
    
    v: list[TagInstance] = []
    async for u in r.users():
        v.append(User(u))
    return TagArray(v)

async def _ctx_waitfor_reply(args: TagArg, kwargs: Kargs, env: Enviroment):
    this = await get_this_msg(env)
    
    r: discord.Message = this.bot.wait_for('message', check=lambda m: this.bot in m.mentions, timeout=100)
    
    if r is None:
        return tag_null
    
    return Message(r)


from .visitor.interpreter import *
from .core.error import *
from .core.klass import *
from .core.node import *
from .visitor.tag_builtin import *
from .visitor import executor
from .visitor.obj_builtin import *
from . import library
from .parser.parser import tparse, TagTransformer


from tagger.core.node import Node
from .tag_builtin import *
from .obj_builtin import *
from . import executor # DO NOT REMOVE. this module initialize all Node execute method 

class TagInterpreter:
    
    def __init__(self) -> None:
        self.globalenv = Enviroment()
        self.globalenv.set('print', TagFunction('print', print_))
        self.globalenv.set('Table', Table)
        self.globalenv.set('slice', TagFunction('slice', slice_))
    
    def start(self, n: Node):
        self.globalenv.error.clear()
        e = n.execute(self.globalenv)
        if not e:
            return
        if e.istype(TagBaseException):
            er = typing.cast(TagException, e)
            self.globalenv.throw(er.error, er.msg, n.l, n.c)
            return
        return e


from tagger.core.node import *
from tagger.core.klass import TagInstance, Enviroment
from .tag_builtin import *

import typing

def _extend(t: type):
    def _(f):
        setattr(t, 'execute', f)
    return _


def evaluate(n: Node, env: Enviroment):
    x = n.execute(env)
    if env.error:
        return tag_null
    if x.istype(TagBaseException):
        y = typing.cast(TagException, x)
        env.throw(y.error, y.msg, n.l, n.c)
        return tag_null
    return x


@_extend(NLiteral)
def _nliteral(node: NLiteral, env: Enviroment):
    if isinstance(node.value, str):
        return TagString(node.value)
    if isinstance(node.value, int):
        return TagInteger(node.value)
    else:
        return TagFloat(node.value)

@_extend(NArithmetic)
def _narith(node: NArithmetic, env: Enviroment):
    l, r = evaluate(node.left, env), evaluate(node.right, env)
    if env.error:
        return tag_null
    
    if node.op == '+':
        return l.add(r)
    elif node.op == '-':
        return l.sub(r)
    elif node.op == '*':
        return l.mult(r)
    elif node.op == '/':
        return l.div(r)

@_extend(NUnary)
def _nunary(node: NUnary, env: Enviroment):
    v = evaluate(node.value, env)
    if env.error:
        return tag_null
    
    if node.op == '+':
        return v.pos()
    elif node.op == '-':
        return v.neg()
    else:
        return v.notb()

@_extend(NSetVar)
def _nsetvar(node: NSetVar, env: Enviroment):
    v = evaluate(node.value, env)
    if env.error:
        return tag_null
    
    env.set(node.name, v)
    return tag_null

@_extend(NCmp)
def _ncmp(node: NCmp, env: Enviroment):
    l, r = evaluate(node.left, env), evaluate(node.right, env)
    if env.error:
        return tag_null
    
    if node.op == '>':
        return l.gt(r)
    elif node.op == '<':
        return l.lt(r)
    elif node.op == '>=':
        return l.gteq(r)
    elif node.op == '<=':
        return l.lteq(r)
    elif node.op == '==':
        return l.equal(r)
    elif node.op == '!=':
        return l.nequal(r)
    elif node.op == 'and':
        return l.andl(r)
    elif node.op == 'or':
        return l.orl(r)

@_extend(NGetVar)
def _ngetvar(node: NGetVar, env: Enviroment):
    v = env.get(node.name)
    if v:
        return v
    env.throw("NameError", f"No {node.name} found in current scope", node.l, node.c)
    return tag_null

@_extend(NDo)
def _ndo(node: NDo, env: Enviroment):
    r: TagInstance = tag_null
    for n in node.tree:
        r = evaluate(n, env)
    return r

@_extend(NDot)
def _ndot(node: NDot, env: Enviroment):
    v = evaluate(node.left, env)
    if env.error:
        return tag_null
    
    x = v.field.get(node.right)
    if x:
        return x
    env.throw('AttributeError', f'No attribute {node.right} found in {v.type}', node.l, node.c)
    return tag_null

@_extend(NFieldSet)
def _nfieldset(node: NFieldSet, env: Enviroment):
    v = evaluate(node.value, env)
    if env.error:
        return tag_null
    
    o = env.get(node.origin)
    if o is None:
        env.throw("NameError", f"No {node.origin} found in current scope", node.l, node.c)
        return tag_null
    
    i = ''
    for n in node.field:
        _o = o
        o = _o.field.get(n)
        if o is None:
            env.throw("AttributeError", f"No attribute {n} found in {v.type}", node.l, node.c)
            return tag_null
        i = n
    
    _o.field[i] = v
    return tag_null

@_extend(NCall)
def _ncall(node: NCall, env: Enviroment):
    c = evaluate(node.caller, env)
    if env.error:
        return
    
    a: TagArg = []
    for arg in node.args:
        a.append(evaluate(arg, env))
    
    ka: Kargs = {}
    for n, arg in node.argk.items():
        ka[n] = evaluate(arg, env)
    
    if env.error:
        return tag_null
    
    return c.call(a, ka, env)

@_extend(NArray)
def _narray(node: NArray, env: Enviroment):
    r = TagArray([])
    for n in node.value:
        a = evaluate(n, env)
        r.value.append(a)
    return r

@_extend(NIfElse)
def _nif(node: NIfElse, env: Enviroment):
    cond = evaluate(node.cond, env)
    if env.error:
        return tag_null
    
    if cond.bool():
        return evaluate(node.body, env)
    else:
        if node.elif_:
            for i in node.elif_:
                c = evaluate(i.cond, env)
                if c.bool():
                    return evaluate(i.body, env)
        if node.else_:
            return evaluate(node.else_.body, env)
    return tag_null

@_extend(NWhile)
def _nwhile(node: NWhile, env: Enviroment):
    r: TagInstance = tag_null
    while evaluate(node.cond, env).bool():
        r = evaluate(node.body, env)
    if env.error:
        return tag_null
    return r

@_extend(NFor)
def _nfor(node: NFor, env: Enviroment):
    v = evaluate(node.value, env)
    if env.error:
        return
    
    s = Enviroment(env)
    for i in v.iter():
        s.set(node.name, i)
        evaluate(node.body, s)
        if s.error:
            return tag_null
    env.error += s.error
    return tag_null


import asyncio
from tagger.core.klass import *
import typing

class TagBase(TagClass):
    NAME = "Base"
    
    def __init__(self) -> None:
        super().__init__()
    
    def add(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        return make_error("TypeError", f"{self.NAME} + {right.type} is not supported")
    def sub(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        return make_error("TypeError", f"{self.NAME} - {right.type} is not supported")
    def mult(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        return make_error("TypeError", f"{self.NAME} * {right.type} is not supported")
    def div(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        return make_error("TypeError", f"{self.NAME} / {right.type} is not supported")
    
    def gt(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        return TagInteger(0)
    def gteq(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        return TagInteger(0)
    def lt(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        return TagInteger(0)
    def lteq(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        return TagInteger(0)
    
    def equal(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        return TagInteger(int(this.field == right.field))
    def nequal(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        return TagInteger(int(not typing.cast(TagInteger, this.equal(right)).bool()))
    
    def pos(self, this: 'TagInstance'):
        return make_error("TypeError", f"{self.NAME} cannot do unary addition operation")
    def neg(self, this: 'TagInstance'):
        return make_error("TypeError", f"{self.NAME} cannot do unary subtration operation")
    def notb(self, this: 'TagInstance'):
        return TagInteger(int(not this.bool()))
    
    def andl(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        return TagInteger(int(this._bool_() and right._bool_()))
    def orl(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        return TagInteger(int(this._bool_() and right._bool_()))
    
    def call(self, this: 'TagInstance', args: TagArg, kargs: Kargs, env: 'Enviroment') -> 'TagInstance':
        return make_error("TypeError", f"{this.type} is not callable")
    
    def iter(self, this: 'TagInstance') -> typing.Generator['TagInstance', typing.Any, None]:
        yield make_error("TypeError", f'{this.type} is not iterable')
    
    def repr(self, this: 'TagInstance') -> 'TagInstance':
        return TagString("[Instance:" + this.type.NAME + "]")
    
    def _bool_(self, this: 'TagInstance'):
        return True
    
    def bool(self, this: 'TagInstance') -> 'TagInstance':
        return TagInteger(int(this._bool_()))

class TagNullType(TagBase):
    NAME = "NullType"
    
    def repr(self, this: 'TagInstance'):
        return TagString('null')
    
    def _bool_(self, this: 'TagInstance') -> bool:
        return False

class TagNull(TagInstance):
    
    def __init__(self) -> None:
        super().__init__()
        self.set_type(tag_nulltype)

class TagBaseException(TagBase):
    NAME = "Exception"

class TagException(TagInstance):
    
    def __init__(self, error: str, msg: str) -> None:
        super().__init__()
        self.error: str = error
        self.msg = msg
        self.set_type(tag_exception)

class TagBaseFunction(TagBase):
    NAME = 'function'
    
    def call(self, this: 'TagInstance', args: TagArg, kargs: Kargs, env: 'Enviroment') -> 'TagInstance':
        t = typing.cast(TagFunction, this)
        t.closure.parent = env
        r = t.body(args, kargs, t.closure)
        t.closure.parent = None
        return r

class TagFunction(TagInstance):
    
    def __init__(self, name: str, body: typing.Callable[[TagArg, Kargs, 'Enviroment'], TagInstance]) -> None:
        super().__init__()
        self.name = name
        self.body = body
        self.closure = Enviroment()
        self.set_type(tag_func)

class TagBaseAFunction(TagBase):
    NAME = 'function'
    
    def call(self, this: 'TagInstance', args: TagArg, kargs: Kargs, env: 'Enviroment') -> 'TagInstance':
        t = typing.cast(TagAFunction, this)
        t.closure.parent = env
        r = asyncio.run(t.body(args, kargs, t.closure))
        t.closure.parent = None
        return r

class TagAFunction(TagInstance):
    
    def __init__(self, name: str, body: typing.Callable[[TagArg, Kargs, 'Enviroment'], typing.Coroutine[typing.Any, typing.Any, TagInstance]]) -> None:
        super().__init__()
        self.name = name
        self.body = body
        self.closure = Enviroment()
        self.set_type(tag_Afunc)

class TagBaseInt(TagBase):
    NAME = 'int'
    
    def add(self, this: TagInstance, right: 'TagInstance'):
        t = typing.cast(TagInteger, this)
        if is_number(right):
            return number(t.value + tnumber(right).value)
        return make_error("TypeError", f"{t.type} + {right.type} is not supported")
    def sub(self, this: TagInstance, right: 'TagInstance'):
        t = typing.cast(TagInteger, this)
        if is_number(right):
            return number(t.value - tnumber(right).value)
        return make_error("TypeError", f"{t.type} + {right.type} is not supported")
    def mult(self, this: TagInstance, right: 'TagInstance'):
        t = typing.cast(TagInteger, this)
        if is_number(right):
            return number(t.value * tnumber(right).value)
        return make_error("TypeError", f"{t.type} + {right.type} is not supported")
    def div(self, this: TagInstance, right: 'TagInstance'):
        t = typing.cast(TagInteger, this)
        if is_number(right):
            r = tnumber(right)
            if r.value == 0:
                return make_error("ZeroDivision", f"Zero division")
            return number(t.value / r.value)
        return make_error("TypeError", f"{t.type} + {right.type} is not supported")
    
    def gt(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        t = typing.cast(TagInteger, this)
        if is_number(right):
            return TagInteger(int(t.value > tnumber(right).value))
        return TagInteger(0)
    def gteq(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        t = typing.cast(TagInteger, this)
        if is_number(right):
            return TagInteger(int(t.value >= tnumber(right).value))
        return TagInteger(0)
    def lt(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        t = typing.cast(TagInteger, this)
        if is_number(right):
            return TagInteger(int(t.value < tnumber(right).value))
        return TagInteger(0)
    def lteq(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        t = typing.cast(TagInteger, this)
        if is_number(right):
            return TagInteger(int(t.value <= tnumber(right).value))
        return TagInteger(0)
    
    def equal(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        if is_number(right):
            return TagInteger(int(typing.cast(TagInteger, this).value == tnumber(right).value))
        return TagInteger(0)
    
    def pos(self, this: TagInstance):
        return number(+(typing.cast(TagInteger, this)).value)
    def neg(self, this: TagInstance):
        return number(-(typing.cast(TagInteger, this)).value)
    
    def _bool_(self, this: 'TagInstance'):
        return typing.cast(TagInteger, this).value != 0
    
    def repr(self, this: 'TagInstance') -> 'TagInstance':
        return TagString(str(typing.cast(TagInteger, this).value))

class TagBaseFloat(TagBase):
    NAME = 'float'
    
    def add(self, this: TagInstance, right: 'TagInstance'):
        t = typing.cast(TagFloat, this)
        if is_number(right):
            return number(t.value + tnumber(right).value)
        return make_error("TypeError", f"{t.type} + {right.type} is not supported")
    def sub(self, this: TagInstance, right: 'TagInstance'):
        t = typing.cast(TagFloat, this)
        if is_number(right):
            return number(t.value - tnumber(right).value)
        return make_error("TypeError", f"{t.type} + {right.type} is not supported")
    def mult(self, this: TagInstance, right: 'TagInstance'):
        t = typing.cast(TagFloat, this)
        if is_number(right):
            return number(t.value * tnumber(right).value)
        return make_error("TypeError", f"{t.type} + {right.type} is not supported")
    def div(self, this: TagInstance, right: 'TagInstance'):
        t = typing.cast(TagFloat, this)
        if is_number(right):
            r = tnumber(right)
            if r.value == 0:
                return make_error("ZeroDivision", f"Zero division")
            return number(t.value / r.value)
        return make_error("TypeError", f"{t.type} + {right.type} is not supported")
    
    def gt(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        t = typing.cast(TagFloat, this)
        if is_number(right):
            return TagInteger(int(t.value > tnumber(right).value))
        return TagInteger(0)
    def gteq(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        t = typing.cast(TagFloat, this)
        if is_number(right):
            return TagInteger(int(t.value >= tnumber(right).value))
        return TagInteger(0)
    def lt(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        t = typing.cast(TagFloat, this)
        if is_number(right):
            return TagInteger(int(t.value < tnumber(right).value))
        return TagInteger(0)
    def lteq(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        t = typing.cast(TagFloat, this)
        if is_number(right):
            return TagInteger(int(t.value <= tnumber(right).value))
        return TagInteger(0)
    
    def equal(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        if is_number(right):
            return TagFloat(int(typing.cast(TagFloat, this).value == tnumber(right).value))
        return TagInteger(0)
    
    def pos(self, this: TagInstance):
        return number(+(typing.cast(TagFloat, this)).value)
    def neg(self, this: TagInstance):
        return number(-(typing.cast(TagFloat, this)).value)
    
    def _bool_(self, this: 'TagInstance'):
        return typing.cast(TagFloat, this).value != 0
    
    def repr(self, this: 'TagInstance') -> 'TagInstance':
        return TagString(str(typing.cast(TagInteger, this).value))

class TagInteger(TagInstance):
    
    def __init__(self, value: int=0) -> None:
        super().__init__()
        self.value = value
        self.set_type(tag_int)
    
    def __repr__(self) -> str:
        return self.value.__repr__()

class TagFloat(TagInstance):
    
    def __init__(self, value: float=0) -> None:
        super().__init__()
        self.value = value
        self.set_type(tag_float)
    
    def __repr__(self) -> str:
        return self.value.__repr__()


class TagBaseString(TagBase):
    NAME = 'string'
    
    def add(self, this: TagInstance, right: 'TagInstance') -> 'TagInstance':
        if right.istype(TagBaseString):
            return TagString(typing.cast(TagString, this).value + typing.cast(TagString, right).value)
        return make_error("TypeError", f"string cannot concatenate with {right.type}")
    
    def mult(self, this: TagInstance, right: 'TagInstance') -> 'TagInstance':
        if right.istype(TagBaseInt):
            return TagString(typing.cast(TagString, this).value * typing.cast(TagInteger, right).value)
        return make_error("TypeError", f"cannot multiply string with {right.type}, must integer")
    
    def equal(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        if right.istype(TagBaseString):
            return TagInteger(int(typing.cast(TagString, this).value == typing.cast(TagString, right).value))
        return TagInteger(0)
    
    def iter(self, this: 'TagInstance') -> typing.Generator['TagInstance', typing.Any, None]:
        for char in typing.cast(TagString, this).value:
            yield TagString(char)
    
    def _bool_(self, this: 'TagInstance'):
        return bool(typing.cast(TagString, this).value)
    
    def repr(self, this: TagInstance):
        return TagString(typing.cast(TagString, this).value.__repr__())

class TagString(TagInstance):
    
    def __init__(self, value: str | list[str] = '') -> None:
        super().__init__()
        if isinstance(value, list):
            self.value = ''.join(value)
        else:
            self.value = value
        self.set_type(tag_str)
        
        add_method(self, 'lower', _str_lower)
        add_method(self, 'upper', _str_upper)
        add_method(self, 'fmt', _str_fmt)
        add_method(self, 'index', _str_index)
    
    def __repr__(self) -> str:
        return self.value

class TagBaseArray(TagBase):
    NAME = 'list'
    
    def add(self, this: TagInstance, right: 'TagInstance') -> 'TagInstance':
        if right.istype(TagBaseArray):
            return TagArray(typing.cast(TagArray, this).value + typing.cast(TagArray, right).value)
        return make_error("TypeError", f"list cannot concatenate with {right.type}")
    
    def mult(self, this: TagInstance, right: 'TagInstance') -> 'TagInstance':
        if right.istype(TagBaseInt):
            return TagArray(typing.cast(TagArray, this).value * typing.cast(TagInteger, right).value)
        return make_error("TypeError", f"cannot multiply list with {right.type}, must integer")
    
    def equal(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        if right.istype(TagBaseArray):
            return TagInteger(int(typing.cast(TagArray, this).value == typing.cast(TagArray, right).value))
        return TagInteger(0)
    
    def iter(self, this: 'TagInstance') -> typing.Generator['TagInstance', typing.Any, None]:
        yield from typing.cast(TagArray, this).value
    
    def _bool_(self, this: 'TagInstance'):
        return bool(typing.cast(TagArray, this).value)
    
    def repr(self, this: TagInstance) -> 'TagInstance':
        return TagString(typing.cast(TagString, this).value.__repr__())

class TagArray(TagInstance):
    
    def __init__(self, value: list[TagInstance]) -> None:
        super().__init__()
        self.value = value
        self.set_type(tag_array)
        
        add_method(self, 'index', _list_index)
        add_method(self, 'contain', _list_contain)
        add_method(self, 'push', _list_push)
        add_method(self, 'remove', _list_remove)
        add_method(self, 'pop', _list_pop)

class TagBaseType(TagBase):
    NAME = 'Type'
    
    def call(self, this: 'TagInstance', args: TagArg, kargs: Kargs, env: 'Enviroment') -> 'TagInstance':
        t = typing.cast(TagType, this)
        o = TagObject(t, env)
        r = o.field['init'].call(args, kargs, o.env)
        if r.istype(TagBaseException):
            return r
        return o

class TagType(TagInstance):
    
    def __init__(self, name: str, methods: dict[str, typing.Callable[[TagArg, Kargs, Enviroment], TagInstance]]) -> None:
        super().__init__()
        self.name = name
        self.methods = methods
        self.set_type(tag_type)
    
    def __repr__(self) -> str:
        return self.name
    
    def method(self, name: str):
        def _(f: typing.Callable[[TagArg, Kargs, Enviroment], TagInstance]):
            self.methods[name] = f
        return _

class TagBaseObject(TagBase):
    NAME = 'Object'
    
    def add(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        t = typing.cast(TagObject, this)
        if 'add' in t.field:
            return t.field['add'].call([right], {}, t.env)
        return make_error('TypeError', f'{t.obj_type}.add is not defined')

    def sub(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        t = typing.cast(TagObject, this)
        if 'sub' in t.field:
            return t.field['add'].call([right], {}, t.env)
        return make_error('TypeError', f'{t.obj_type}.sub is not defined')

    def mult(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        t = typing.cast(TagObject, this)
        if 'mult' in t.field:
            return t.field['add'].call([right], {}, t.env)
        return make_error('TypeError', f'{t.obj_type}.mult is not defined')

    def div(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        t = typing.cast(TagObject, this)
        if 'div' in t.field:
            return t.field['add'].call([right], {}, t.env)
        return make_error('TypeError', f'{t.obj_type}.div is not defined')

    def gt(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        t = typing.cast(TagObject, this)
        if 'gt' in t.field:
            return t.field['add'].call([right], {}, t.env)
        return make_error('TypeError', f'{t.obj_type}.gt is not defined')

    def gteq(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        t = typing.cast(TagObject, this)
        if 'gteq' in t.field:
            return t.field['add'].call([right], {}, t.env)
        return make_error('TypeError', f'{t.obj_type}.gteq is not defined')

    def lt(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        t = typing.cast(TagObject, this)
        if 'lt' in t.field:
            return t.field['add'].call([right], {}, t.env)
        return make_error('TypeError', f'{t.obj_type}.lt is not defined')

    def lteq(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        t = typing.cast(TagObject, this)
        if 'lteq' in t.field:
            return t.field['add'].call([right], {}, t.env)
        return make_error('TypeError', f'{t.obj_type}.lteq is not defined')

    def equal(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        t = typing.cast(TagObject, this)
        if 'equal' in t.field:
            return t.field['add'].call([right], {}, t.env)
        return make_error('TypeError', f'{t.obj_type}.equal is not defined')

    def nequal(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        t = typing.cast(TagObject, this)
        if 'nequal' in t.field:
            return t.field['add'].call([right], {}, t.env)
        return make_error('TypeError', f'{t.obj_type}.nequal is not defined')

    def andl(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        t = typing.cast(TagObject, this)
        if 'andl' in t.field:
            return t.field['add'].call([right], {}, t.env)
        return make_error('TypeError', f'{t.obj_type}.andl is not defined')

    def orl(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance':
        t = typing.cast(TagObject, this)
        if 'orl' in t.field:
            return t.field['add'].call([right], {}, t.env)
        return make_error('TypeError', f'{t.obj_type}.orl is not defined')
    
    def pos(self, this: 'TagInstance'):
        t = typing.cast(TagObject, this)
        if 'pos' in t.field:
            return t.field['positive'].call([], {}, t.env)
        return make_error('TypeError', f'{t.obj_type}.positive is not defined')
    
    def neg(self, this: 'TagInstance'):
        t = typing.cast(TagObject, this)
        if 'neg' in t.field:
            return t.field['negate'].call([], {}, t.env)
        return make_error('TypeError', f'{t.obj_type}.negate is not defined')
    
    def notb(self, this: 'TagInstance'):
        t = typing.cast(TagObject, this)
        if 'not' in t.field:
            return t.field['notb'].call([], {}, t.env)
        return make_error('TypeError', f'{t.obj_type}.notb is not defined')
    
    def call(self, this: 'TagInstance', args: TagArg, kargs: Kargs, env: 'Enviroment') -> 'TagInstance':
        t = typing.cast(TagObject, this)
        if 'call' in t.field:
            return t.field['call'].call([], {}, t.env)
        return make_error('TypeError', f'{t.obj_type} is not callable')
    
    def repr(self, this: 'TagInstance'):
        t = typing.cast(TagObject, this)
        return t.field['repr'].call([], {}, t.env)

class TagObject(TagInstance):
    
    def __init__(self, type: TagType, env: Enviroment) -> None:
        super().__init__()
        self.obj_type = type
        self.env = Enviroment(env)
        self.env.set('this', self)
        self.sfield: dict[str, TagInstance] = {}
        self.set_type(tag_object)
        
        add_method(self, 'init', _empty_method)
        add_method(self, 'repr', _obj_repr)
        
        for n, m in self.obj_type.methods.items():
            add_method(self, n, m)

def tnumber(v: TagInstance):
    if v.istype(TagBaseInt):
        return typing.cast(TagInteger, v)
    elif v.istype(TagBaseFloat):
        return typing.cast(TagFloat, v)
    raise Exception(f"tnumber got non-number, {v.type}")

def number(v: int | float):
    if isinstance(v, int):
        return TagInteger(v)
    return TagFloat(v)

def is_number(v: TagInstance):
    return v.istype(TagBaseInt) or v.istype(TagBaseFloat)

tag_exception = TagBaseException()
tag_int = TagBaseInt()
tag_float = TagBaseFloat()
tag_str = TagBaseString()
tag_func = TagBaseFunction()
tag_Afunc = TagBaseAFunction()
tag_nulltype = TagNullType()
tag_array = TagBaseArray()
tag_type = TagBaseType()
tag_object = TagBaseObject()

tag_null = TagNull()

def make_error(err: str, msg: str):
    return TagException(err, msg)

def argerror(where: str, need: int, got: list[typing.Any]):
    return make_error("TypeError", f"{where} required {need} argument, got {len(got)}")

def add_method(t: TagInstance, name: str, func: typing.Callable[[TagArg, Kargs, 'Enviroment'], TagInstance]):
    m = TagFunction(name, func)
    m.closure.set('this', t)
    t.field[name] = m

def add_Amethod(t: TagInstance, name: str, func: typing.Callable[[TagArg, Kargs, 'Enviroment'], typing.Coroutine[typing.Any, typing.Any, TagInstance]]):
    m = TagAFunction(name, func)
    m.closure.set('this', t)
    t.field[name] = m

def _str_lower(args: TagArg, kwargs: Kargs, env: Enviroment):
    this = typing.cast(TagString, env.get('this'))
    return TagString(this.value.lower())

def _str_upper(args: TagArg, kwargs: Kargs, env: Enviroment):
    this = typing.cast(TagString, env.get('this'))
    return TagString(this.value.upper())

def _str_fmt(args: TagArg, kwargs: Kargs, env: Enviroment):
    this = typing.cast(TagString, env.get('this'))
    
    s = this.value
    for i, a in enumerate(args):
        _a = a.__repr__()
        s = s.replace('{}', _a, 1)
        if '{'+str(i)+'}' in _a:
            s = s.replace('{'+str(i)+'}', _a)
    
    for n, v in kwargs.items():
        s = s.replace('{'+n+'}', v.__repr__())
    
    return TagString(s)

def _list_index(args: TagArg, kwargs: Kargs, env: Enviroment):
    if len(args) != 1:
        return argerror('list.index', 1, args)
    i = args[0]
    
    if i.istype(TagBaseInt):
        try:
            return typing.cast(TagArray, env.get('this')).value[typing.cast(TagInteger, i).value]
        except IndexError:
            return make_error('IndexError', 'list out of bound')
    return make_error('TypeError', f'list indexing only accept integer not {i.type}')

def _list_contain(args: TagArg, kwargs: Kargs, env: Enviroment):
    if len(args) != 1:
        return argerror('list.contain', 1, args)
    i = args[0]

    for o in env.get('this').iter():
        if i.equal(o)._bool_():
            return TagInteger(1)
    return TagInteger(0)

def _list_remove(args: TagArg, kwargs: Kargs, env: Enviroment):
    if len(args) != 1:
        return argerror('list.remove', 1, args)
    i = args[0]

    this = typing.cast(TagArray, env.get('this'))
    
    for c, o in enumerate(this.iter()):
        if i.equal(o)._bool_():
            this.value.pop(c)
    
    return tag_null

def _list_push(args: TagArg, kwargs: Kargs, env: Enviroment):
    if len(args) != 1:
        return argerror('list.append', 1, args)
    i = args[0]

    this = typing.cast(TagArray, env.get('this'))
    
    this.value.append(i)
    
    return tag_null

def _list_pop(args: TagArg, kwargs: Kargs, env: Enviroment):
    this = typing.cast(TagArray, env.get('this'))
    
    if len(args) != 1:
        try:
            this.value.pop()
        except IndexError:
            return make_error('IndexError', 'list out of bound')
    
    i = args[0]
    
    if not i.istype(TagBaseInt):
        return make_error('TypeError', f'list.pop required integer not {i.type}')

    try:
        this.value.pop(typing.cast(TagInteger, i).value)
    except IndexError:
        return make_error('IndexError', 'list out of bound')
    
    return tag_null

def _str_index(args: TagArg, kwargs: Kargs, env: Enviroment):
    if len(args) != 1:
        return argerror('str.index', 1, args)
    i = args[0]
    
    if i.istype(TagBaseInt):
        try:
            return TagString(typing.cast(TagString, env.get('this')).value[typing.cast(TagInteger, i).value])
        except IndexError:
            return make_error('IndexError', 'string out of bound')
    return make_error('TypeError', f'string indexing only accept integer not {i.type}')

def _empty_method(args: TagArg, kwargs: Kargs, env: Enviroment):
    return tag_null

def _obj_repr(args: TagArg, kwargs: Kargs, env: Enviroment):
    this = typing.cast(TagObject, env.get('this'))
    return TagString(f'[Instance:{this.obj_type.name}]')

def get_this(env: Enviroment):
    return typing.cast(TagObject, env.get('this'))

def icopy(i: TagInstance, *args: typing.Any):
    s = i.__class__(*args)
    s.field = {**i.field.copy(), **s.field}
    return s


import random

from tagger.visitor.obj_builtin import  *

def randint(args: TagArg, kargs: Kargs, env: Enviroment):
    success, er = expect('randint', [TagBaseInt, TagBaseInt], {}, args, kargs)
    if not success:
        return er
    
    a_, b_ = cast(TagInteger, args[0]).value, cast(TagInteger, args[1]).value
    
    return TagInteger(random.randint(a_, b_))

def randchoice(args: TagArg, kargs: Kargs, env: Enviroment):
    success, er = expect('randchoice', [TagArray], {}, args, kargs)
    if not success:
        return er
    
    a = cast(TagArray, args[0])
    
    if len(a.value) == 0:
        return tag_null
    
    return random.choice(a.value)


from typing import Optional
import typing
from .error import Error

TagArg = list['TagInstance']
Kargs = dict[str, 'TagInstance']

class TagClass:
    NAME = "BaseClass"
    
    # def new(self):
    #     t = TagInstance()
    #     t.type = self
    #     return t
    
    def repr(self, this: 'TagInstance') -> 'TagInstance': # type: ignore
        pass
    
    def add(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance': # type: ignore
        pass
    def sub(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance': # type: ignore
        pass
    def mult(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance': # type: ignore
        pass
    def div(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance': # type: ignore
        pass
    
    def pos(self, this: 'TagInstance') -> 'TagInstance': # type: ignore
        pass
    def neg(self, this: 'TagInstance') -> 'TagInstance': # type: ignore
        pass
    def notb(self, this: 'TagInstance') -> 'TagInstance': # type: ignore
        pass
    
    def gt(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance': # type: ignore
        pass
    def lt(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance': # type: ignore
        pass
    def gteq(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance': # type: ignore
        pass
    def lteq(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance': # type: ignore
        pass
    def equal(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance': # type: ignore
        pass
    def nequal(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance': # type: ignore
        pass
    def andl(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance': # type: ignore
        pass
    def orl(self, this: 'TagInstance', right: 'TagInstance') -> 'TagInstance': # type: ignore
        pass
    
    def bool(self, this: 'TagInstance') -> 'TagInstance': # type: ignore
        pass
    
    def call(self, this: 'TagInstance', args: TagArg, kargs: Kargs, env: 'Enviroment') -> 'TagInstance': # type: ignore
        pass
    
    def iter(self, this: 'TagInstance') -> typing.Generator['TagInstance', typing.Any, None]: # type: ignore
        pass
    
    def __repr__(self) -> str:
        return self.NAME
    
    def _bool_(self, this: 'TagInstance') -> 'bool': # type: ignore
        return True

class TagInstance:
    
    def __init__(self) -> None:
        self.type: TagClass
        self.field: dict[str, TagInstance] = {}
    
    def set_type(self, t: TagClass):
        self.type = t
    
    def istype(self, klass: type):
        return isinstance(self.type, klass)
    
    def add(self, right: 'TagInstance') -> 'TagInstance':
        return self.type.add(self, right)
    def sub(self, right: 'TagInstance') -> 'TagInstance':
        return self.type.sub(self, right)
    def mult(self, right: 'TagInstance') -> 'TagInstance':
        return self.type.mult(self, right)
    def div(self, right: 'TagInstance') -> 'TagInstance':
        return self.type.div(self, right)

    def gt(self, right: 'TagInstance') -> 'TagInstance':
        return self.type.gt(self, right)
    def gteq(self, right: 'TagInstance') -> 'TagInstance':
        return self.type.gteq(self, right)

    def lt(self, right: 'TagInstance') -> 'TagInstance':
        return self.type.lt(self, right)
    def lteq(self, right: 'TagInstance') -> 'TagInstance':
        return self.type.lteq(self, right)

    def equal(self, right: 'TagInstance') -> 'TagInstance':
        return self.type.equal(self, right)
    def nequal(self, right: 'TagInstance') -> 'TagInstance':
        return self.type.nequal(self, right)

    def andl(self, right: 'TagInstance') -> 'TagInstance':
        return self.type.andl(self, right)
    def orl(self, right: 'TagInstance') -> 'TagInstance':
        return self.type.orl(self, right)
    
    def bool(self) -> 'TagInstance':
        return self.type.bool(self)
    
    def _bool_(self) -> 'bool': # type: ignore
        return self.type._bool_(self)
    
    def pos(self) -> 'TagInstance':
        return self.type.pos(self)
    def neg(self) -> 'TagInstance':
        return self.type.neg(self)
    def notb(self) -> 'TagInstance':
        return self.type.notb(self)
    
    def call(self, args: TagArg, kargs: Kargs, env: 'Enviroment') -> 'TagInstance':
        return self.type.call(self, args, kargs, env)
    
    def iter(self) -> typing.Generator['TagInstance', typing.Any, None]:
        yield from self.type.iter(self)
    
    def repr(self) -> 'TagInstance':
        return self.type.repr(self)
    
    def __repr__(self) -> str:
        return self.repr().__repr__()


class Enviroment:
    
    def __init__(self, parent: Optional['Enviroment']=None) -> None:
        self.parent = parent
        self.table: dict[str, TagInstance] = {}
        self.error: list[Error] = []
    
    def throw(self, err: str, msg: str, l: int, c: int):
        self.error.append(Error(err, msg, l, c))

    def get(self, n: str):
        if n not in self.table:
            if self.parent:
                return self.parent.get(n)
        return self.table.get(n)
    
    def set(self, n: str, v: TagInstance):
        self.table[n] = v
    
    def __iter__(self):
        yield from self.table.items()


from typing import Optional
from .klass import TagInstance, Enviroment
from dataclasses import dataclass

@dataclass
class Node:
    l: int
    c: int
    
    def execute(self, env: Enviroment) -> TagInstance: # type: ignore
        pass

@dataclass
class NLiteral(Node):
    value: str | int | float

@dataclass
class NArray(Node):
    value: list[Node]

@dataclass
class NArithmetic(Node):
    left: Node
    op: str
    right: Node

@dataclass
class NUnary(Node):
    op: str
    value: Node

@dataclass
class NCmp(Node):
    left: Node
    op: str
    right: Node

@dataclass
class NDo(Node):
    tree: list[Node]

@dataclass
class NSetVar(Node):
    name: str
    value: Node

@dataclass
class NGetVar(Node):
    name: str

@dataclass
class NDot(Node):
    left: Node
    right: str

@dataclass
class NCall(Node):
    caller: Node
    args: list[Node]
    argk: dict[str, Node]

@dataclass
class NFieldSet(Node):
    origin: str
    field: list[str]
    value: Node

@dataclass
class NIfElse(Node):
    cond: Node
    body: Node
    elif_: list['NElif']
    else_: Optional['NElse']

@dataclass
class NElif(Node):
    cond: Node
    body: Node

@dataclass
class NElse(Node):
    body: Node

@dataclass
class NWhile(Node):
    cond: Node
    body: Node

@dataclass
class NFor(Node):
    name: str
    value: Node
    body: Node


from .tag_builtin import *
from typing import cast

def tstring(t: TagInstance):
    return typing.cast(TagString, t).value

def tarray(t: TagInstance):
    return typing.cast(TagArray, t).value


def print_(args: TagArg, kwargs: Kargs, env: Enviroment):
    sep = kwargs.get('sep')
    if sep is None:
        _sep = ' '
    elif sep.istype(TagBaseString):
        _sep = typing.cast(TagString, sep).value
    else:
        _sep = ' '
    
    s = ''
    for arg in args:
        r = arg.__repr__()
        if isinstance(r, str):
            s += r + _sep
        else:
            return make_error('TypeError', f'{arg.type}.repr did not return string')
    print(s)
    return tag_null

Table = TagType('Table', {})
_LOCKED_METHOD_TABLE = ('init','get','set')

@Table.method('init')
def table_init(args: TagArg, kwargs: Kargs, env: Enviroment):
    this = get_this(env)
    
    for n, k in kwargs.items():
        if n in _LOCKED_METHOD_TABLE:
            return make_error('TableError', f'{n} is not allowed name for table item')
        this.field[n] = k
    
    return tag_null

@Table.method('get')
def table_get(args: TagArg, kwargs: Kargs, env: Enviroment):
    this = get_this(env)
    
    if len(args) != 1:
        return argerror('Table.get', 1, args)
    
    a = args[0]
    if a.istype(TagBaseString):
        return this.field[typing.cast(TagString, a).value]
    return make_error('TypeError', f'Table.get required string not {a.type}')

@Table.method('set')
def table_set(args: TagArg, kwargs: Kargs, env: Enviroment):
    this = get_this(env)
    
    if len(args) != 2:
        return argerror('Table.set', 2, args)
    
    n = args[0]
    v = args[1]
    
    if n.istype(TagBaseString):
        this.field[typing.cast(TagString, n).value] = v
        return this
    return make_error('TypeError', f'Table.set first argument required string not {n.type}')

@Table.method('repr')
def table_repr(args: TagArg, kwargs: Kargs, env: Enviroment):
    this = get_this(env)
    return TagString(this.field.__repr__())


def slice_(args: TagArg, kwargs: Kargs, env: Enviroment):
    if len(args) != 1:
        return argerror('slice', 1, args)
    v = args[0]
    
    s, e = kwargs.get('s') or tag_null, kwargs.get('e') or tag_null
    
    if s.istype(TagBaseInt):
        _s = typing.cast(TagInteger, s).value
    else:
        _s = None
    if e.istype(TagBaseInt):
        _e = typing.cast(TagInteger, e).value
    else:
        _e = None
    
    g = list(v.iter())
    if g:
        if g[0].istype(TagBaseException):
            return g
    
    try:
        s_ = g[_s:_e]
    except IndexError:
        return make_error("IndexError", f"{v.type} out of bound")
    
    if v.istype(TagBaseString):
        return TagString([typing.cast(TagString, x).value for x in s_])
    return TagArray(s_)

def expect(where: str, args: list[type], kwargs: dict[str, type], a: TagArg, k: Kargs):
    if len(a) != len(args):
        return False, argerror(where, len(args), a)
    
    _i = 1
    for arg, argt in zip(a, args):
        if not arg.istype(argt):
            return False, make_error('TypeError', f'{where} argument {_i} expected to be {argt} but got {arg}')
        _i += 1
    
    for n, v in k.items():
        if n in kwargs:
            if not v.istype(kwargs[n]):
                return False, make_error('TypeError', f'{where} argument \'{n}\' expected to be {kwargs[n]} but got {v.type}')
    
    for n in kwargs.keys():
        if n not in k:
            return False, make_error('TypeError', f'{where} required \'{n}\' argument')
    
    return True, None


from lark import Lark, Transformer, Token, Tree
from lark.exceptions import UnexpectedCharacters, UnexpectedEOF, UnexpectedToken
from tagger.core.node import *
from tagger.core.error import Error

tparser = Lark(open('parser/tagger.lark'))

def tparse(text: str):
    try:
        return tparser.parse(text)
    except UnexpectedCharacters as e:
        return Error('SyntaxError', 'Unexpected character', e.line, e.column)
    except UnexpectedToken as e:
        return Error('SyntaxError', 'Invalid Syntax', e.line, e.column)
    except UnexpectedEOF:
        l = text.splitlines()
        return Error('SyntaxError', 'Unexpted End of line', len(l), len(l[-1]))

Terminal = tuple[Token]
BinOp = tuple[Node, Token, Node]

class TagTransformer(Transformer):
    
    def start(self, t):
        return t

    def integer(self, t: Terminal):
        (v,) = t
        return NLiteral(v.line, v.column, int(v)) # type: ignore
    
    def float_(self, t: Terminal):
        (v,) = t
        return NLiteral(v.line, v.column, float(v)) # type: ignore
    
    def string(self, t: Terminal):
        (v,) = t
        a = str(v[1:-1])
        _a = ''
        _p = ''
        for c in a:
            if _p == '\\':
                if c == 'n':
                    _a += '\n'
                elif c == 't':
                    _a += '\t'
                elif c == 'r':
                    _a += '\r'
                else:
                    _a += c
                _p = c
                continue
            _p = c
            if c != '\\':
                _a += c
        return NLiteral(v.line, v.column, _a) # type: ignore
    
    def array(self, t):
        b, v = t[0], t[1:]
        return NArray(b.line, b.column, v or []) # type: ignore
    
    def arith(self, t: BinOp):
        l, o, r = t
        return NArithmetic(o.line, o.column, l, o, r) # type: ignore
    
    def unary(self, t: tuple[Token, Node]):
        o, v = t
        return NUnary(o.line, o.column, o, v) # type: ignore
    
    def cmp(self, t: BinOp):
        l, o, r = t
        return NCmp(o.line, o.column, l, o, r) # type: ignore
    
    def tagdo(self, t: list[Node]):
        d, a = t[0], t[1:]
        return NDo(d.line, d.column, a) # type: ignore
    
    def tagvar(self, t: tuple[Token, Node]):
        n, v = t
        return NSetVar(n.line, n.column, str(n), v) # type: ignore
    
    def getvar(self, t: Terminal):
        (n,) = t
        return NGetVar(n.line, n.column, str(n)) # type: ignore
    
    def dot(self, t: BinOp):
        l, o, r = t
        return NDot(o.line, o.column, l, r) # type: ignore
    
    def keyarg(self, t: tuple[Token, Node]):
        return (str(t[0]), t[1])
    
    def call(self, t: tuple[Node, list[Node | tuple[str, Node]]]):
        caller, arg = t
        r = NCall(caller.l, caller.c, caller, [], {}) # type: ignore
        for a in arg.children: # type: ignore
            if isinstance(a, Node):
                r.args.append(a)
            else:
                r.argk[a[0]] = a[1]
        return r
    
    def fieldset(self, t: tuple[list[Token], Node]):
        n, v = t[0].children, t[1] # type: ignore
        return NFieldSet(n[0].line, n[0].column, n[0], n[1:], v) # type: ignore
    
    # syntantic sugar
    def indexing(self, t: tuple[Node, Token, NLiteral]):
        n, o, i = t
        return NCall(o.line, o.column, NDot(o.line, o.column, n, 'index'), [i], {}) # type: ignore
    
    def tagelse(self, t: tuple[Token, Node]):
        s, b = t
        if s is None:
            return None
        return NElse(s.line, s.column, b)
    
    def tagelif(self, t: tuple[Token, Node, Node]):
        s, c, b = t
        if s is None:
            return None
        return NElif(s.line, s.column, c, b)
    
    def tagif(self, t: list[Token | Tree | Node]): # type: ignore
        stmt, cond, body = t[:3]
        elif_ = t[3:-1]
        else_ = t[-1]
        
        if None in elif_:
            elif_ = []
        
        return NIfElse(stmt.line, stmt.column, cond, body, elif_, else_) # type: ignore
    
    def tagwhile(self, t: tuple[Token, Node, Node]):
        s, c, b = t
        return NWhile(s.line, s.column, c, b)
    
    def tagfor(self, t: tuple[Token, Token, Node, Node]):
        s, n, v, b = t
        return NFor(s.line, s.column, str(n), v, b)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tagger', 'tagger.core', 'tagger.library', 'tagger.parser', 'tagger.visitor']

package_data = \
{'': ['*']}

install_requires = \
['lark>=1.1.5,<2.0.0']

setup_kwargs = {
    'name': 'taggerscript',
    'version': '0.0.9',
    'description': 'Scripting language written in Python for Discord bot',
    'long_description': '# Tagger\n\nTagger is a scripting language for Discord bot. It can be used from tag command to user automation moderation. It is simple but powerful. Users cannot run code that can harm your bot. You can add new function and class to it in a simple way.\n\n# Example\n\n```tagger\n{do\n    {msg.set("React with :heart: to claim this present")}\n    {msg.wait_for_reaction(":heart:")}\n    {msg.set("Just kidding there is no present")}\n}\n```\nCheck example for more\n\n# Install\n\n## PIP\n```\npip install taggerscript\n```\n\n## Poetry\n```\npoetry add taggerscript\n```\n\n## git clone\n```\ngit clone https://github.com/najis-poop/TaggerScript\n```\n\n# Documentation\n\ncoming soon\n\n# Feedback\n\nTagger still in development. Any bug should be reported quickly by creating issue in github issue.',
    'author': 'najis',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass

@dataclass
class Error:
    err: str
    msg: str
    l: int
    c: int
    
    def __repr__(self) -> str:
        return f"Error at line {self.l}, column {self.c}\n{self.err}: {self.msg}"